"""
Functions to retrieve data that are specific to areas and relevant to all indicators. For example: Deprivation decile.
"""


import pandas as pd
import warnings
from fingertips_py.retrieve_data import get_data_by_indicator_ids

def defined_qcut(df, value_series, number_of_bins, bins_for_extras, labels=False):
    """
    Allows users to define how values are split into bins when clustering.

    :param df: Dataframe of values
    :param value_series: Name of value column to rank upon
    :param number_of_bins: Integer of number of bins to create
    :param bins_for_extras: Ordered list of bin numbers to assign uneven splits
    :param labels: Optional. Labels for bins if required
    :return: A dataframe with a new column 'bins' which contains the cluster numbers
    """
    if max(bins_for_extras) > number_of_bins or any(x < 0 for x in bins_for_extras):
        raise ValueError('Attempted to allocate to a bin that doesnt exist')
    base_number, number_of_values_to_allocate = divmod(df[value_series].count(), number_of_bins)
    bins_for_extras = bins_for_extras[:number_of_values_to_allocate]
    if number_of_values_to_allocate == 0:
        df['bins'] = pd.qcut(df[value_series], number_of_bins, labels=labels)
        return df
    elif number_of_values_to_allocate > len(bins_for_extras):
        raise ValueError('There are more values to allocate than the list provided, please select more bins')
    bins = {}
    for i in range(number_of_bins):
        number_of_values_in_bin = base_number
        if i in bins_for_extras:
            number_of_values_in_bin += 1
        bins[i] = number_of_values_in_bin
    df['rank'] = df[value_series].rank()
    df = df.sort_values(by=['rank'])
    df['bins'] = 0
    row_to_start_allocate = 0
    row_to_end_allocate = 0
    for bin_number, number_in_bin in bins.items():
        row_to_end_allocate += number_in_bin
        bins.update({bin_number: [number_in_bin, row_to_start_allocate, row_to_end_allocate]})
        row_to_start_allocate = row_to_end_allocate
    conditions = [df['rank'].iloc[v[1]: v[2]] for k, v in bins.items()]
    series_to_add = pd.Series()
    for idx, series in enumerate(conditions):
        series[series > -1] = idx
        series_to_add = series_to_add.append(series)
    df['bins'] = series_to_add
    df['bins'] = df['bins'] + 1
    df = df.reset_index()
    return df


extra_areas = {
    1: [0],
    2: [0, 5],
    3: [0, 3, 7],
    4: [0, 2, 6, 8],
    5: [0, 2, 4, 6, 8],
    6: [0, 1, 3, 5, 6, 8],
    7: [0, 1, 2, 4, 5, 7, 8],
    8: [0, 1, 2, 3, 5, 6, 7, 8],
    9: [0, 1, 2, 3, 4, 5, 6, 7, 8]
}


def deprivation_decile(area_type_id, year='2015', area_code=None):
    """
    Takes in an area type id and returns a pandas series of deprivation deciles for those areas (with the areas as an
    index. If a specific area is requested, it returns just the deprivation decile value.

    :param area_type_id: Area type id as denoted by the Fingertips API
    :param year: Year of deprivation score
    :param area_code: Optional. Area code for area type to return a single value for that area
    :return: A pandas series of deprivation scores with area codes as the index. Or single value if area is specified.
    """
    warnings.warn('Caution, the deprivation deciles are being calculated on the fly and might show some inconsistencies'
                  ' from the live Fingertips site.')
    acceptable_deprivation_years_la = ['2010', '2015']
    acceptable_deprivation_years_gp = ['2015']
    acceptable_area_types = [3, 101, 102, 7, 153]
    order_of_extra_values = []
    if not isinstance(year, str):
        year = str(year)
    if year not in acceptable_deprivation_years_la and area_type_id != 7:
        raise ValueError \
            ('The acceptable years are 2010 and 2015 for local authorities and CCGs, please select one of these')
    elif year not in acceptable_deprivation_years_gp:
        raise ValueError('The acceptable years are 2015, please select this')
    if area_type_id not in acceptable_area_types:
        raise ValueError('Currently, we support deprivation decile for District & UA, County & UA, MSOA and GP area '
                         'types')
    if area_type_id == 3:
        indicator_id = 93275
        area_dep_dec = get_data_by_indicator_ids(indicator_id, area_type_id, parent_area_type_id=101, profile_id=143,
                                                 include_sortable_time_periods=True)
    else:
        indicator_id = 91872
        area_dep_dec = get_data_by_indicator_ids(indicator_id, area_type_id)
    if area_type_id == 102:
        order_of_extra_values = [0, 9, 1, 2, 3, 4, 5, 6, 7, 8]

    area_dep_dec = area_dep_dec[area_dep_dec['Area Code'] != 'E92000001']
    if not order_of_extra_values:
        order_of_extra_values = extra_areas[area_dep_dec['Value'].count() % 10]
    area_dep_dec = defined_qcut(area_dep_dec, 'Value', 10, order_of_extra_values)
    area_dep_dec.set_index('Area Code', inplace=True)
    if area_code:
        try:
            return area_dep_dec.loc[area_code, 'decile']
        except KeyError:
            raise KeyError('This area is not available at in this area type. Please try another area type')
    return area_dep_dec['bins']






import pandas as pd
import pytest
from fingertips_py.api_calls import get_json, make_request, get_json_return_df, base_url
from fingertips_py.retrieve_data import get_all_data_for_profile, get_all_data_for_indicators, get_data_by_indicator_ids, \
    get_all_areas_for_all_indicators, get_data_for_indicator_at_all_available_geographies
from fingertips_py.metadata import get_metadata_for_profile_as_dataframe, get_metadata, get_metadata_for_indicator_as_dataframe, \
    get_metadata_for_domain_as_dataframe, get_all_value_notes, get_profile_by_name, get_area_types_for_profile, \
    get_domains_in_profile, get_all_profiles, get_area_types_as_dict, get_age_from_id, get_area_type_ids_for_profile, \
    get_profile_by_id, get_age_id, get_all_ages, get_all_sexes, get_areas_for_area_type, get_metadata_for_indicator, \
    get_multiplier_and_calculation_for_indicator, get_sex_from_id, get_sex_id, get_value_note_id, \
    get_metadata_for_all_indicators, get_metadata_for_all_indicators_from_csv, get_all_areas, get_profile_by_key
from fingertips_py.area_data import deprivation_decile


def test_get_json():
    data = get_json('https://jsonplaceholder.typicode.com/todos/1')
    assert data['userId'] == 1
    assert isinstance(data, dict)

def test_make_request():
    data = make_request(base_url + 'area_types', 'Id')
    assert type(data) == dict
    assert data[15]['Name'] == 'England'


def test_get_json_return_df():
    data = get_json_return_df('https://jsonplaceholder.typicode.com/todos/1')
    assert isinstance(data, pd.DataFrame) is True


# need to think about this one
def test_get_all_data_for_profile():
    data = get_all_data_for_profile(84, area_type_id=402, is_test=True)
    assert isinstance(data[0], pd.DataFrame) is True
    assert data[1] == base_url + 'all_data/csv/by_profile_id?child_area_type_id=402&parent_area_type_id=15&profile_id=84'


def test_get_all_areas():
    test_dict, url = get_all_areas(is_test=True)
    first_dict = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert isinstance(first_dict, dict) is True
    assert isinstance(first_dict.get('Name'), str) is True
    assert url == 'https://fingertips.phe.org.uk/api/area_types'


def test_get_all_data_for_indicators():
    data, url = get_all_data_for_indicators([92949, 92998], area_type_id=102, is_test=True)
    assert isinstance(data, pd.DataFrame) is True
    assert data.shape[1] == 27
    assert url == 'https://fingertips.phe.org.uk/api/all_data/csv/by_indicator_id?indicator_ids=92949,92998&child_area_type_id=102&parent_area_type_id=15'


def test_get_data_by_indicator_ids():
    data, url = get_data_by_indicator_ids([92949, 92998], 102, is_test=True)
    assert isinstance(data, pd.DataFrame) is True
    assert data.shape[1] == 27
    assert url == 'https://fingertips.phe.org.uk/api/all_data/csv/by_indicator_id?indicator_ids=92949,92998&child_area_type_id=102&parent_area_type_id=15'


def test_get_all_areas_for_all_indicators():
    test_dict = get_all_areas_for_all_indicators()
    first_list = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert 15 in first_list


def test_get_data_for_indicator_at_all_available_geographies():
    data = get_data_for_indicator_at_all_available_geographies(92998)
    assert isinstance(data, pd.DataFrame) is True
    assert data.shape[1] == 27


def test_get_metadata_for_profile_as_dataframe():
    data = get_metadata_for_profile_as_dataframe(84)
    assert isinstance(data, pd.DataFrame) is True
    assert data.shape[1] == 32


def test_get_metadata():
    data_indicators = get_metadata(indicator_ids=[92949, 90581])
    data_domain = get_metadata(domain_ids=[1938133052, 1938132811])
    data_profile = get_metadata(profile_ids=84)
    data_indicators_and_domain = get_metadata(indicator_ids=[92949, 90581], domain_ids=[1938133052, 1938132811])
    data_domain_and_profile = get_metadata(domain_ids=[1938133052, 1938132811], profile_ids=84)
    #data_profile_and_indicators = get_metadata(indicator_ids=[92949, 90581], profile_ids=84) ####!
    data_all = get_metadata(indicator_ids=[92949, 90581], domain_ids=[1938133052, 1938132811], profile_ids=84)
    assert isinstance(data_indicators, pd.DataFrame) is True
    assert data_indicators.shape[1] == 32
    assert isinstance(data_domain, pd.DataFrame) is True
    assert data_domain.shape[1] == 32
    assert isinstance(data_profile, pd.DataFrame) is True
    assert data_profile.shape[1] == 32
    assert isinstance(data_indicators_and_domain, pd.DataFrame) is True
    assert data_indicators_and_domain.shape[1] == 32
    assert isinstance(data_domain_and_profile, pd.DataFrame) is True
    assert data_domain_and_profile.shape[1] == 32
    #assert isinstance(data_profile_and_indicators, pd.DataFrame) is True
    #assert data_profile_and_indicators.shape[1] == 32
    assert isinstance(data_all, pd.DataFrame) is True
    assert data_all.shape[1] == 32


def test_get_metadata_for_indicator_as_dataframe():
    data = get_metadata_for_indicator_as_dataframe(247, is_test=True)
    assert isinstance(data[0], pd.DataFrame) is True
    assert data[0].shape[1] == 32
    assert data[1] == 'https://fingertips.phe.org.uk/api/indicator_metadata/csv/by_indicator_id?indicator_ids=247'


def test_get_metadata_for_domain_as_dataframe():
    data = get_metadata_for_domain_as_dataframe(1938132811, is_test=True)
    assert isinstance(data[0], pd.DataFrame) is True
    assert data[0].shape[1] == 32
    assert data[1] == 'https://fingertips.phe.org.uk/api/indicator_metadata/csv/by_group_id?group_id=1938132811'


def test_get_all_value_notes():
    test_dict = get_all_value_notes()
    first_item = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert isinstance(first_item, str) is True


def test_get_profile_by_name():
    data = get_profile_by_name('dementia')
    assert isinstance(data, dict) is True
    assert data['Id'] == 84


def test_get_profile_by_key():
    data = get_profile_by_key("general-practice")
    assert isinstance(data, dict) is True
    assert data['Id'] == 20

# no
def test_get_area_types_for_profile():
    test_dict = get_area_types_for_profile(84)
    first_item = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert isinstance(first_item, dict) is True


def test_get_domains_in_profile():
    test_list = get_domains_in_profile(84)
    assert isinstance(test_list, list) is True
    assert isinstance(test_list[0], int) is True


def test_get_all_profiles():
    test_dict = get_all_profiles()
    first_item = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert isinstance(first_item, dict) is True


def test_get_area_types_as_dict():
    test_dict = get_area_types_as_dict()
    first_item = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert isinstance(first_item, dict) is True


def test_get_age_from_id():
    data = get_age_from_id(1)
    assert data == 'All ages'


def test_get_area_type_ids_for_profile():
    data = get_area_type_ids_for_profile(84)
    assert isinstance(data, list) is True
    assert isinstance(data[1], int) is True


def test_get_profile_by_id():
    data = get_profile_by_id(84)
    assert isinstance(data, dict) is True


def test_get_age_id():
    data = get_age_id('All ages')
    assert data == 1


def test_get_all_ages():
    test_dict = get_all_ages()
    first_item = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert isinstance(first_item, dict) is True


def test_get_all_sexes():
    test_dict = get_all_sexes()
    first_item = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert isinstance(first_item, str) is True


def test_get_areas_for_area_type():
    data = get_areas_for_area_type(102)
    assert isinstance(data, dict) is True
    assert isinstance(list(data.values())[0], dict) is True


def test_get_metadata_for_indicator():
    data = get_metadata_for_indicator(247)
    assert isinstance(data, dict) is True


def test_get_multiplier_and_calculation_for_indicator():
    data = get_multiplier_and_calculation_for_indicator(247)
    assert isinstance(data, tuple) is True
    assert data[0] == 100


def test_get_sex_from_id():
    data = get_sex_from_id(1)
    assert data == 'Male'


def test_get_sex_id():
    data = get_sex_id('Male')
    assert data == 1


def test_get_value_note_id():
    data = get_value_note_id('Value suppressed for disclosure control due to small count')
    assert data == 101


def test_get_metadata_for_all_indicators():
    test_dict = get_metadata_for_all_indicators()
    first_item = test_dict.get(next(iter(test_dict)))
    assert isinstance(test_dict, dict) is True
    assert len(first_item.keys()) == 15


def test_get_metadata_for_all_indicators_from_csv():
    data = get_metadata_for_all_indicators_from_csv()
    assert isinstance(data, pd.DataFrame) is True
    assert data.shape[1] == 32

# deprivation decile needs to work
def test_deprivation_decile():
    data = deprivation_decile(7)
    assert len(data.unique()) == 10



"""
Calls used to retrieve metadata about areas, ages, sexes, value notes, calculation methods, rates, and indicator
metadata.
"""

import pandas as pd
from urllib.error import HTTPError, URLError
from fingertips_py.api_calls import base_url, make_request, get_json, get_json_return_df, deal_with_url_error, get_data_in_dict


def get_all_ages(is_test=False):
    """
    Returns a dictionary of all the age categories and their IDs as the dictionary key.

    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: Age codes used in Fingertips in a dictionary
    """
    ages = get_data_in_dict(base_url + 'ages')
    if is_test:
        return ages, base_url + 'ages'
    return ages


def get_all_areas(is_test=False):
    """
    Retreives all area types.

    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dictionary of all area types used in Fingertips
    """
    areas = make_request(base_url + 'area_types', 'Id')
    if is_test:
        return areas, base_url + 'area_types'
    return areas


def get_age_id(age, is_test=False):
    """
    Returns an ID for a given age.

    :param age: Search term of an age or age range as a string
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: Code used in Fingertips to represent the age as an integer or age range as a string
    """
    ages = make_request(base_url + 'ages', 'Name')
    if is_test:
        return ages[age]['Id'], base_url + 'ages'
    return ages[age]['Id']


def get_age_from_id(age_id, is_test=False):
    """
    Returns an age name from given id.

    :param age_id: Age id used in Fingertips as an integer
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: Age, or age range, as a string
    """
    ages = make_request(base_url + 'ages', 'Id')
    if is_test:
        return ages[age_id]['Name'], base_url + 'ages'
    return ages[age_id]['Name']


def get_all_sexes(is_test=False):
    """
    Returns a dictionary of all sex categories and their IDs as dictionary key.

    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: Sex categories used in Fingertips with associated codes as a dictionary
    """
    sexes = get_data_in_dict(base_url + 'sexes', value = 'Name')
    if is_test:
        return sexes, base_url + 'sexes'
    return sexes


def get_sex_id(sex, is_test=False):
    """
    Returns an ID for a given sex.

    :param sex: Sex category as string (Case sensitive)
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: ID used in Fingertips to represent the sex as an integer
    """
    sexes = make_request(base_url + 'sexes', 'Name')
    if is_test:
        return sexes[sex]['Id'], base_url + 'sexes'
    return sexes[sex]['Id']


def get_sex_from_id(sex_id, is_test=False):
    """
    Returns a sex name given an ID.

    :param sex_id: ID used in Fingertips to represent the sex as integer
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: Sex category as string
    """
    sexes = make_request(base_url + 'sexes', 'Id')
    if is_test:
        return sexes[sex_id]['Name'], base_url + 'sexes'
    return sexes[sex_id]['Name']


def get_all_value_notes(is_test=False):
    """
    Returns a dictionary of all value notes and their IDs as dictionary key.

    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: Data value notes and their associated codes that are used in Fingertips as a dictionary
    """
    value_notes = get_data_in_dict(base_url + 'value_notes', value = 'Text')
    if is_test:
        return value_notes, base_url + 'value_notes'
    return value_notes


def get_value_note_id(value_note, is_test=False):
    """
    Returns a value note ID for a given value note.

    :param value_note: Value note as string
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: ID used in Fingertips to represent the value note as an integer
    """
    value_notes = make_request(base_url + 'value_notes', 'Text')
    if is_test:
        return value_notes[value_note]['Id'], base_url + 'value_notes'
    return value_notes[value_note]['Id']


def get_areas_for_area_type(area_type_id, is_test=False):
    """
    Returns a dictionary of areas that match an area type id given the id as integer or string.

    :param area_type_id: ID of area type (ID of General Practice is 7 etc) used in Fingertips as integer or string
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dictionary of dictionaries with area codes as the key 
    """
    areas = make_request(base_url + 'areas/by_area_type?area_type_id=' + str(area_type_id), 'Code')
    if is_test:
        return areas, base_url + 'areas/by_area_type?area_type_id=' + str(area_type_id)
    return areas


def get_metadata_for_indicator(indicator_number, is_test=False):
    """
    Returns the metadata for an indicator given the indicator number as integer or string.

    :param indicator_number: Number used to identify an indicator within Fingertips as integer or string
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dictionary of metadata for the given indicator
    """
    metadata = get_json(base_url + 'indicator_metadata/by_indicator_id?indicator_ids=' + str(indicator_number))
    metadata_dict = metadata.get(str(indicator_number))
    if is_test:
        return metadata, base_url + 'indicator_metadata/by_indicator_id?indicator_ids=' + str(indicator_number)
    return metadata_dict


def get_metadata_for_all_indicators_from_csv(is_test=False):
    """
    Returns a dataframe from the csv of all metadata for all indicators.

    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dataframe of all metadata for all indicators
    """
    try:
        metadata = pd.read_csv(base_url + 'indicator_metadata/csv/all')
    except URLError:
        metadata = deal_with_url_error(base_url + 'indicator_metadata/csv/all')
    if is_test:
        return metadata, base_url + 'indicator_metadata/csv/all'
    return metadata


def get_metadata_for_all_indicators(include_definition='no', include_system_content='no', is_test=False):
    """
    Returns the metadata for all indicators in a dictionary.

    :param include_definition: optional to include definitions
    :param include_system_content: optional to include system content
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: dictionary of all indicators
    """
    url_suffix = f'indicator_metadata/all?include_definition={include_definition}&include_system_content={include_system_content}'
    metadata_dict = get_json(base_url + url_suffix)
    if is_test:
        return metadata_dict, base_url + url_suffix
    return metadata_dict


def get_multiplier_and_calculation_for_indicator(indicator_number):
    """
    Returns the multiplier and calculation method for a given indicator.

    :param indicator_number: Number used to identify an indicator within Fingertips as integer or string
    :return: A tuple of multiplier and calculation method from Fingetips metadata
    """
    metadata = get_metadata_for_indicator(indicator_number)
    multiplier = metadata.get('Unit').get('Value')
    calc_metadata = metadata.get('ConfidenceIntervalMethod').get('Name')
    if 'wilson' in calc_metadata.lower():
        calc = 'Wilson'
    elif 'byar' in calc_metadata.lower():
        calc = 'Byar'
    else:
        calc = None
    return multiplier, calc


def get_area_types_as_dict(is_test=False):
    """
    Returns all area types and related information such as ID and name with dictionary key value as ID.

    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dictionary of area types
    """
    areas = get_data_in_dict(base_url + 'area_types')
    if is_test:
        return areas, base_url + 'area_types'
    return areas


def get_profile_by_id(profile_id, is_test=False):
    """
    Returns a profile as an dictionary which contains information about domains and sequencing.

    :param profile_id: ID used in Fingertips to identify a profile as integer or string
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dictionary of information about the profile
    """
    if is_test:
        return get_json(base_url + 'profile?profile_id=' + str(profile_id)), base_url + 'profile?profile_id=' + \
               str(profile_id)
    return get_json(base_url + 'profile?profile_id=' + str(profile_id))


def get_all_profiles(is_test=False):
    """
    Returns all profiles.

    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dictionary of all profiles in Fingertips including information on domains and sequencing
    """
    profiles = get_data_in_dict(base_url + 'profiles')
    if is_test:
        return profiles, base_url + 'profiles'
    return profiles


def get_domains_in_profile(profile_id):
    """
    Returns the domain IDs for a given profile.

    :param profile_id: ID used in Fingertips to identify a profile as integer or string
    :return: A list of domain IDs
    """
    profile = get_profile_by_id(profile_id)
    return profile['GroupIds']


def get_area_types_for_profile(profile_id, is_test=False):
    """
    Retrieves all the area types that have data for a given profile.

    :param profile_id: ID used in Fingertips to identify a profile as integer or string
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A list of dictionaries of area types with relevant information
    """
    if is_test:
        return get_data_in_dict(base_url + 'area_types?profile_ids=' + str(profile_id)), base_url + 'area_types?profile_ids=' + \
               str(profile_id)
    return get_data_in_dict(base_url + 'area_types?profile_ids=' + str(profile_id))


def get_area_type_ids_for_profile(profile_id):
    """
    Returns a list of area types used within a given profile.

    :param profile_id: ID used in Fingertips to identify a profile as integer or string
    :return: A list of area type IDs used within a given profile
    """
    area_type_obj = get_area_types_for_profile(profile_id)
    area_type_list = [value.get('Id') for value in area_type_obj.values()]
    return area_type_list


def get_profile_by_name(profile_name):
    """
    Returns a profile object given a name to search – try to be specific to get better results.

    :param profile_name: A string or part of a string that is used as the profile name
    :return: A dictionary of the profile metadata including domain information or an error message
    """
    all_profiles = get_all_profiles()
    profile_obj = ''
    for profile in all_profiles.values():
        if profile_name.lower() in profile.get('Name').lower():
            profile_obj = profile
    if not profile_obj:
        return 'Profile could not be found'
    else:
        return profile_obj


def get_profile_by_key(profile_key):
    """
    Returns a profile object given a key (as the stub following 'profile' in the website URL). For example,
    give, a URL of the form `https://fingertips.phe.org.uk/profile/general-practice/data#page/3/gid/2000...`,
    the key is 'general-practice'.

    :param profile_key: The exact key for the profile.
    :return: A dictionary of the profile metadata including domain information or an error message
    """
    all_profiles = get_all_profiles()
    for profile_id, profile_object in all_profiles.items():
        if profile_object.get('Key') == profile_key:
            return profile_object
    return 'Profile could not be found'


def get_metadata_for_indicator_as_dataframe(indicator_ids, is_test=False):
    """
    Returns a dataframe of metadata for a given indicator ID or list of indicator IDs.

    :param indicator_ids: Number or list of numbers used to identify an indicator within Fingertips as integer or string
    :return: Dataframe object with metadate for the indicator ID
    """
    url_suffix = "indicator_metadata/csv/by_indicator_id?indicator_ids={}"
    if isinstance(indicator_ids, list):
        indicator_ids = ','.join(list(map(str, indicator_ids)))
    try:
        df = pd.read_csv(base_url + url_suffix.format(str(indicator_ids)))
    except HTTPError:
        raise NameError(f'Indicator {indicator_ids} does not exist')
    except URLError:
        df = deal_with_url_error(base_url + url_suffix.format(str(indicator_ids)))
    if is_test:
        return df, base_url + url_suffix.format(str(indicator_ids))
    return df


def get_metadata_for_domain_as_dataframe(group_ids, is_test=False):
    """
    Returns a dataframe of metadata for a given domain ID or list of domain IDs.

    :param group_ids: Number or list of numbers used to identify a domain within Fingertips as integer or string
    :return: Dataframe object with metadata for the indicators for a given domain ID
    """
    url_suffix = "indicator_metadata/csv/by_group_id?group_id={}"
    if isinstance(group_ids, list):
        df = pd.DataFrame()
        for group_id in group_ids:
            try:
                df = pd.concat([df, pd.read_csv(base_url + url_suffix.format(str(group_id)))])
            except HTTPError:
                raise NameError(f'Domain {group_id} does not exist')
            except URLError:
                df = deal_with_url_error(base_url + url_suffix.format(str(group_id)))
    else:
        try:
            df = pd.read_csv(base_url + url_suffix.format(str(group_ids)))
        except HTTPError:
            raise NameError(f'Domain {group_ids} does not exist')
        except URLError:
            df = deal_with_url_error(base_url + url_suffix.format(str(group_ids)))
    if is_test:
        return df, base_url + url_suffix.format(str(group_ids))
    return df


def get_metadata_for_profile_as_dataframe(profile_ids):
    """
    Returns a dataframe of metadata for a given profile ID or list of profile IDs.

    :param profile_ids: ID or list of IDs used in Fingertips to identify a profile as integer or string
    :return: Dataframe object with metadata for the indicators for a given group ID
    """
    url_suffix = "indicator_metadata/csv/by_profile_id?profile_id={}"
    if isinstance(profile_ids, list):
        df = pd.DataFrame()
        for profile_id in profile_ids:
            try:
                df = pd.concat([df, pd.read_csv(base_url + url_suffix.format(str(profile_id)))])
            except HTTPError:
                raise NameError(f'Profile {profile_id} does not exist')
            except URLError:
                df = deal_with_url_error(base_url + url_suffix.format(str(profile_id)))
    else:
        try:
            df = pd.read_csv(base_url + url_suffix.format(str(profile_ids)))
        except HTTPError:
            raise NameError(f'Profile {profile_ids} does not exist')
        except URLError:
            df = deal_with_url_error(base_url + url_suffix.format(str(profile_ids)))
    return df


def get_metadata(indicator_ids=None, domain_ids=None, profile_ids=None):
    """
    Returns a dataframe object of metadata for a given indicator, domain, and/or profile given the relevant IDs. At
    least one of these IDs has to be given otherwise an error is raised.

    :param indicator_ids: [OPTIONAL] Number used to identify an indicator within Fingertips as integer or string
    :param domain_ids: [OPTIONAL] Number used to identify a domain within Fingertips as integer or string
    :param profile_ids: [OPTIONAL] ID used in Fingertips to identify a profile as integer or string
    :return: A dataframe object with metadata for the given IDs or an error if nothing is specified
    """
    if indicator_ids and domain_ids and profile_ids:
        df = get_metadata_for_profile_as_dataframe(profile_ids)
        df = pd.concat([df, get_metadata_for_domain_as_dataframe(domain_ids)])
        df = pd.concat([df, get_metadata_for_indicator_as_dataframe(indicator_ids)])
        return df
    if indicator_ids and domain_ids:
        df = get_metadata_for_domain_as_dataframe(domain_ids)
        df = pd.concat([df, get_metadata_for_indicator_as_dataframe(indicator_ids)])
        return df
    if indicator_ids and profile_ids:
        df = get_metadata_for_profile_as_dataframe(profile_ids)
        df = pd.concat([df, get_metadata_for_indicator_as_dataframe(indicator_ids)])
        return df
    if domain_ids and profile_ids:
        df = get_metadata_for_profile_as_dataframe(profile_ids)
        df = pd.concat([df, get_metadata_for_domain_as_dataframe(domain_ids)])
        return df
    if profile_ids:
        return get_metadata_for_profile_as_dataframe(profile_ids)
    if domain_ids:
        return get_metadata_for_domain_as_dataframe(domain_ids)
    if indicator_ids:
        return get_metadata_for_indicator_as_dataframe(indicator_ids)
    raise NameError('Must use a valid indicator IDs, domain IDs or profile IDs')



"""
A group of functions to retrieve data from Fingertips by indicator, profile, domain (group), or geography.
"""


import pandas as pd
from urllib.error import URLError, HTTPError
from fingertips_py.api_calls import base_url, deal_with_url_error, get_json
from fingertips_py.metadata import get_area_type_ids_for_profile, get_metadata_for_all_indicators, get_all_areas


def get_data_by_indicator_ids(indicator_ids, area_type_id, parent_area_type_id=15, profile_id=None,
                              include_sortable_time_periods=None, is_test=False):
    """
    Returns a dataframe of indicator data given a list of indicators and area types.
    :param indicator_ids: Single indicator ID or list of indicator IDs, as integers or strings
    :param area_type_id: ID of area type (eg. CCG, Upper Tier Local Authority) used in Fingertips as integer or string
    :param parent_area_type_id: Area type of parent area - defaults to England value
    :param profile_id: ID of profile to select by as either int or string
    :param include_sortable_time_periods: Boolean as to whether to include a sort-friendly data field
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dataframe of data relating to the given indicators
    """

    url_suffix = 'all_data/csv/by_indicator_id?indicator_ids={}&child_area_type_id={}&parent_area_type_id={}'
    if profile_id and not include_sortable_time_periods:
        url_addition = f'&profile_id={profile_id}'
        url_suffix = url_suffix + url_addition
    elif include_sortable_time_periods and not profile_id:
        url_addition = '&include_sortable_time_periods=yes'
        url_suffix = url_suffix + url_addition
    elif profile_id and include_sortable_time_periods:
        url_addition = f'&profile_id={profile_id}&include_sortable_time_periods=yes'
        url_suffix = url_suffix + url_addition
    if isinstance(indicator_ids, list):
        if any(isinstance(ind, int) for ind in indicator_ids):
            indicator_ids = ','.join(str(ind) for ind in indicator_ids)
        else:
            indicator_ids = ','.join(indicator_ids)
    else:
        indicator_ids = str(indicator_ids)
    populated_url = url_suffix.format(indicator_ids, str(area_type_id), parent_area_type_id)
    try:
        df = pd.read_csv(base_url + populated_url, low_memory = False)
    except URLError:
        df = deal_with_url_error(base_url + populated_url)
    if is_test:
        return df, base_url + populated_url
    return df


def get_all_data_for_profile(profile_id, parent_area_type_id=15, area_type_id = None, filter_by_area_codes=None,
                             is_test=False):
    """
    Returns a dataframe of data for all indicators within a profile.

    :param profile_id: ID used in Fingertips to identify a profile as integer or string
    :param parent_area_type_id: Area type of parent area - defaults to England value
    :param area_type_id: Option to only return data for a given area type. Area type ids are string, int or a list.
    :param filter_by_area_codes: Option to limit returned data to areas. Areas as either string or list of strings.
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: A dataframe of data for all indicators within a profile with any filters applied
    """
    if area_type_id is not None:
        if type(area_type_id) == int:
            area_types = [area_type_id]
        else:
            area_types = area_type_id
    else:
        area_types = get_area_type_ids_for_profile(profile_id)
    df = pd.DataFrame()
    for area in area_types:
        populated_url = (f'all_data/csv/by_profile_id?child_area_type_id={area}&'
                         f'parent_area_type_id={parent_area_type_id}&'
                         f'profile_id={profile_id}')
        try:
            df_returned = pd.read_csv(base_url + populated_url, low_memory = False)
        except HTTPError:
            raise Exception('There has been a server error with Fingertips for this request. ')
        except URLError:
            df_returned = deal_with_url_error(base_url + populated_url)
        df = pd.concat([df, df_returned])
    if filter_by_area_codes:
        if isinstance(filter_by_area_codes, list):
            df = df.loc[df['Area Code'].isin(filter_by_area_codes)]
        elif isinstance(filter_by_area_codes, str):
            df = df.loc[df['Area Code'] == filter_by_area_codes]
        df = df.reset_index()
    if is_test:
        return df, base_url + populated_url
    return df

def get_all_data_for_indicators(indicators, area_type_id, parent_area_type_id=15, filter_by_area_codes=None,
                                is_test=False):
    """
    Returns a dataframe of data for given indicators at an area.

    :param indicators: List or integer or string of indicator IDs
    :param area_type_id: ID of area type (eg. ID of General Practice is 7 etc) used in Fingertips as integer or string
    :param parent_area_type_id: Area type of parent area - defaults to England value
    :param filter_by_area_codes: Option to limit returned data to areas. Areas as either string or list of strings
    :param is_test: Used for testing. Returns a tuple of expected return and the URL called to retrieve the data
    :return: Dataframe of data for given indicators at an area
    """
    if isinstance(indicators, list):
        if any(isinstance(ind, int) for ind in indicators):
            indicators = ','.join(str(ind) for ind in indicators)
        else:
            indicators = ','.join(indicators)
    else:
        indicators = str(indicators)
        
    populated_url = (f'all_data/csv/by_indicator_id?indicator_ids={indicators}&'
                     f'child_area_type_id={area_type_id}&'
                     f'parent_area_type_id={parent_area_type_id}')
    try:
        df = pd.read_csv(base_url + populated_url, low_memory = False)
    except URLError:
        df = deal_with_url_error(base_url + populated_url)
    df.reset_index()
    if filter_by_area_codes:
        if isinstance(filter_by_area_codes, list):
            df = df.loc[df['Area Code'].isin(filter_by_area_codes)]
        elif isinstance(filter_by_area_codes, str):
            df = df.loc[df['Area Code'] == filter_by_area_codes]
        df = df.reset_index()
    if is_test:
        return df, base_url + populated_url
    return df


def get_all_areas_for_all_indicators():
    """
    Returns a dictionary of all indicators and their geographical breakdowns.

    :return: Dictionary of all indicators (ID as key) and their geographical breakdowns
    """
    url_suffix = 'available_data'
    all_area_ids = get_json(base_url + url_suffix)
    all_indicators = list(set([x.get('IndicatorId') for x in all_area_ids]))
    all_indicators.sort()
    area_dict = {}
    for ind in all_indicators:
        area_list = []
        for item in all_area_ids:
            if item.get('IndicatorId') == ind:
                area_list.append(item.get('AreaTypeId'))
        area_dict[ind] = area_list
    return area_dict
                

def get_data_for_indicator_at_all_available_geographies(indicator_id):
    """
    Returns a dataframe of all data for an indicator for all available geographies.

    :param indicator_id: Indicator id
    :return: Dataframe of data for indicator for all available areas for all time periods
    """
    all_area_for_all_indicators = get_all_areas_for_all_indicators()
    areas_to_get = all_area_for_all_indicators.get(indicator_id)
    df = pd.DataFrame()
    for area in areas_to_get:
        df_temp = get_data_by_indicator_ids(indicator_id, area)
        df = pd.concat([df, df_temp])
    df.drop_duplicates(inplace=True)
    return df




__version__ = '0.4.1'

from fingertips_py.api_calls import get_json, make_request
from fingertips_py.retrieve_data import get_all_data_for_profile, get_all_data_for_indicators, get_data_by_indicator_ids, \
    get_all_areas_for_all_indicators, get_data_for_indicator_at_all_available_geographies
from fingertips_py.metadata import get_metadata_for_profile_as_dataframe, get_metadata, get_metadata_for_indicator_as_dataframe, \
    get_metadata_for_domain_as_dataframe, get_all_value_notes, get_profile_by_name, get_area_types_for_profile,\
    get_domains_in_profile, get_all_profiles, get_area_types_as_dict, get_age_from_id, get_area_type_ids_for_profile, \
    get_profile_by_id, get_age_id, get_all_ages, get_all_sexes, get_areas_for_area_type, get_metadata_for_indicator, \
    get_multiplier_and_calculation_for_indicator, get_sex_from_id, get_sex_id, get_value_note_id, \
    get_metadata_for_all_indicators, get_metadata_for_all_indicators_from_csv, get_all_areas
from fingertips_py.area_data import deprivation_decile



"""
A group of functions to query the Fingertips api and retrieve data in a variety of formats.
"""


import requests
import json
import pandas as pd
from io import StringIO


def make_request(url, attr=None):
    """
    :param url: A url to make a request
    :param attr: The attribute that needs to be returned
    :return: a dict of the attribute and associated data

    """
    try:
        req = requests.get(url)
    except requests.exceptions.SSLError:
        req = requests.get(url, verify=False)
    json_response = json.loads(req.content.decode('utf-8'))
    data = {}
    for item in json_response:
        name = item.pop(attr)
        data[name] = item
    return data


def get_json(url):
    """
    :param url: A url to make a request
    :return: A parsed JSON object
    """
    try:
        req = requests.get(url)
    except requests.exceptions.SSLError:
        req = requests.get(url, verify=False)
    json_resp = json.loads(req.content.decode('utf-8'))
    return json_resp


def get_json_return_df(url, transpose=True):
    """
    :param url: A url to make a request
    :param transpose: [OPTIONAL] transposes dataframe. Default True.
    :return: Dataframe generated from JSON response.

    :meta private:
    """
    try:
        req = requests.get(url)
    except requests.exceptions.SSLError:
        req = requests.get(url, verify=False)
    try:
        df = pd.read_json(req.content, encoding='utf-8')
    except TypeError:
        df = pd.DataFrame.from_dict([req.json()])
    if transpose:
        df = df.transpose()
    return df
    
    
def get_data_in_dict(url, key = None, value = None):
    """
    :param url: A url to make a request
    :param key: The item in the JSON to be used as the dictionary key
    :param value: The item in the JSON to be used as the dictionary value
    :return: A dictionary of returned data using first item as dictionary key by default

    :meta private:
    """
    json_list = get_json(url)
    if key is None:
        key = list(json_list[0].keys())[0]
    json_dict = {}
    if value is None:
        for js in json_list:
            json_dict[js.get(key)] = js
    else:
        for js in json_list:
            json_dict[js.get(key)] = js.get(value)
    return json_dict


def deal_with_url_error(url):
    """
    :param url: A url that returns a URL Error based on SSL errors
    :return: A dataframe from the URL with varify set to false.

    :meta private:
    """
    req = requests.get(url, verify=False)
    s = str(req.content, 'utf-8')
    data = StringIO(s)
    df = pd.read_csv(data)
    return df


base_url = 'https://fingertips.phe.org.uk/api/'


